-------------------------------------------------------
-- Memory module for Warren's microcoded CPU.
--
-- There are four banks of 16K words. The first bank starting at 0x0000
-- contains ROM. The second bank at 0x4000 contains RAM. The third bank
-- is empty. The fourth bank at 0xC000 maps I/O devices like the UART.
--
-- As per the original design, the memory only requires an address bus,
-- a data bus, a clock and a load (write enable) line.
--
-- When the load line is enabled (high), the data bus goes high impedance.
-- (c) 2015 Warren Toomey, GPL3 licence
-------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity memory is
    generic (ADDR_WIDTH :    integer := 16;
             DATA_WIDTH :    integer := 16
    );
    port (clk     :in    std_logic;
          address :in    std_logic_vector (ADDR_WIDTH-1 downto 0);
          data    :inout std_logic_vector (DATA_WIDTH-1 downto 0);
          ld      :in    std_logic
    );
end entity;

architecture behaviour of memory is
    -- Internal signal lines
    -- The top two bits of the incoming address
    alias addrsel: std_logic_vector (1 downto 0)
      			is address(ADDR_WIDTH-1 downto ADDR_WIDTH-2);
    
    -- The lower bits of the incoming address
    alias addrlsb: std_logic_vector (ADDR_WIDTH-3 downto 0)
      			is address(ADDR_WIDTH-3 downto 0);
    
    -- Test lines just for debugging purposes, not used
    signal test_addrsel: std_logic_vector (1 downto 0);
    signal test_addrlsb: std_logic_vector (ADDR_WIDTH-3 downto 0);
    
    -- We want RAM and ROM chip select always, and always output enable
    signal cs: std_logic := '1';
    signal oe: std_logic := '1';
    
    -- The load line for RAM and its output
    signal ram_ld:   std_logic;
    signal ram_data: std_logic_vector (DATA_WIDTH-1 downto 0);
    
    -- The ROM's output
    signal rom_data: std_logic_vector (DATA_WIDTH-1 downto 0);
    
    -- The UART is mapped as follows:
    --    data_out is at 0xC000 lowest 8 bits
    --    data_in is at 0xC001 lowest 8 bits
    --    dav,ready,empty are the lowest 3 bits at 0xC002
    signal uart_data_out: std_logic_vector(7 downto 0);
    signal uart_data_in: std_logic_vector(7 downto 0);
    signal uart_outbits: std_logic_vector(2 downto 0);
    signal uart_read: std_logic;
    signal uart_write: std_logic;
    
    -- The 16K ROM component
    component rom is
        generic (filename : string :="main.rom";
                 ADDR_WIDTH :integer := ADDR_WIDTH-2;
            	 DATA_WIDTH :integer := DATA_WIDTH
        );
        port (ce      :in  std_logic;
            read_en :in  std_logic;
            address :in  std_logic_vector (ADDR_WIDTH-1 downto 0);
            data    :out std_logic_vector (DATA_WIDTH-1 downto 0)
          );
    end component;
    
    -- The 16K RAM component
    component ram is
        generic (ADDR_WIDTH :integer := ADDR_WIDTH-2;
            	 DATA_WIDTH :integer := DATA_WIDTH
        );
        port (
            clk     :in    std_logic;
            address :in    std_logic_vector (ADDR_WIDTH-1 downto 0);
            data    :inout std_logic_vector (DATA_WIDTH-1 downto 0);
            cs      :in    std_logic;
            we      :in    std_logic;
            oe      :in    std_logic);
    end component;
    
    -- The UART component
    component ghdl_uart
        port (
            data_out : out std_logic_vector(7 downto 0);
            data_in  : in  std_logic_vector(7 downto 0);
            dav      : out std_logic;   -- received data available
            ready    : out std_logic;	-- there is free space in tx buffer
            empty    : out std_logic;   -- the transmit buffer is empty
            rd       : in  std_logic;   -- asynchronous read strobe
            wr       : in  std_logic    -- asynchronous write strobe
        );
    end component;
    
begin
    -- The ROM's port map
    romchip: rom port map (
        address => addrlsb,
        data => rom_data,
        ce => cs,
        read_en => oe
      );
    
    -- The RAM's port map
    ramchip: ram port map (
        clk => clk,
        address => addrlsb,
        data => ram_data,
        cs => cs,
        we => ram_ld,
        oe => oe
      );
    
    -- The UART's port map
    uartchip: ghdl_uart port map (
        data_out => uart_data_out,
        data_in => uart_data_in,
        dav => uart_outbits(2),
        ready => uart_outbits(1),
        empty => uart_outbits(0),
        rd => uart_read,
        wr => uart_write
      );
    
    -- The data is: high impedance when ld is high,
    -- the ROM output when address bank 0,
    -- the RAM output when address bank 1,
    -- the UART's data output when address 0xC000,
    -- the UART's status lines when address 0xC002
    -- and (for now) high impedance for everything else
    data <= (others => 'Z') when ld='1' else
      	    rom_data when addrsel= "00" else
      	    ram_data when addrsel= "01" else
      	    "00000000" & uart_data_out when address= x"C000" else
      	    "0000000000000" & uart_outbits when address= x"C002" else
      	    (others => 'Z');
    
    -- RAM is only written when using address bank 1.
    -- We only send data to RAM when ld is asserted and
    -- when using address bank 1
    ram_ld <= ld when addrsel= "01" else
      	      '0';
    ram_data <= data when (ld='1' and addrsel= "01") else
      	  	(others => 'Z');
    
    -- We send data to the UART when we write to 0xC001
    uart_data_in <= data(7 downto 0) when address= x"C001";
    
    -- We strobe the UART read line when the address is 0xC000.
    -- Ditto we strobe the write line on address 0xC001.
    uart_read  <= '1' when address= x"C000" else '0';
    uart_write <= '1' when address= x"C001" and ld='1' else '0';
    
    -- Debugging
    test_addrsel <= addrsel;
    test_addrlsb <= addrlsb;
end architecture;

library ieee;
use ieee.std_logic_1164.all;

entity memory_test is
    generic (DATA_WIDTH :integer := 16;
             ADDR_WIDTH :integer := 16
    );
end memory_test;

architecture behaviour of memory_test is
    component memory is
        generic (ADDR_WIDTH :    integer := 16;
                 DATA_WIDTH :    integer := 16
        );
        port (clk     :in    std_logic;
              address :in    std_logic_vector (ADDR_WIDTH-1 downto 0);
              data    :inout std_logic_vector (DATA_WIDTH-1 downto 0);
              ld      :in    std_logic
        );
    end component;
    
    signal clock  : std_logic;
    signal addr   : std_logic_vector(ADDR_WIDTH-1 downto 0) := x"0000";
    signal data   : std_logic_vector(DATA_WIDTH-1 downto 0);
    signal ld     : std_logic;
    
    -- Clock period definitions
    constant clock_period : time := 10 ns;
    
begin
    -- Clock process definitions: clock with 50% duty cycle is generated here.
    clock_process: process
    begin
        clock <= '0';
        wait for clock_period/2;  --for 5 ns signal is '0'.
        clock <= '1';
        wait for clock_period/2;  --for next 5 ns signal is '1'.
    end process;
    
    -- Unit under test port map
    uut: memory port map (
        clk => clock,
        address => addr,
        data => data,
        ld => ld
      );
    
    --- Stimulation process
    stim_proc: process
    begin
        --        -- Read from RAM location 0x4017
        --        ld <= '0'; 
        --        addr <= x"4017"; wait for clock_period;
        --        data <= (others => 'Z');
        --        
        --        -- Store at RAM location 0x4022
        --        ld <= '1'; 
        --        addr <= x"4022";
        --        data <= x"ab"; wait for clock_period;
        --        
        --        -- Store ee at RAM location 0x4088
        --        ld <= '1'; 
        --        addr <= x"4080";
        --        data <= x"ee"; wait for clock_period;
        --        
        --        -- Read from RAM location 0x4022
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"4022"; wait for clock_period;
        --        
        --        -- Read from some other location
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"4020"; wait for clock_period;
        --        
        --        -- Read from 0x4022 again
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"4022"; wait for clock_period;
        --        
        --        -- Read from 0x4080 again
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"4080"; wait for clock_period;
        --        wait for clock_period;
        --
        --	-- Read from ROM location 0x0001
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"0001"; wait for clock_period;
        --        wait for clock_period;
        --        
        --	-- Read from ROM location 0x0002
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"0002"; wait for clock_period;
        --        wait for clock_period;
        --        
        --	-- Read from ROM location 0x0020
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"0020"; wait for clock_period;
        --        wait for clock_period;
        --
        --	-- UART testing
        wait for clock_period/2;
        ld <= '1'; 
        addr <= x"C001";
        data <= x"0042"; wait for clock_period;
        ld <= '0'; wait for clock_period;
        
        ld <= '1'; 
        addr <= x"C001";
        data <= x"0048"; wait for clock_period;
        ld <= '0'; wait for clock_period;
        
        ld <= '1'; 
        addr <= x"C001";
        data <= x"0046"; wait for clock_period;
        ld <= '0'; wait for clock_period;
        
        --	-- Write the letter 'A' to the output port at 0xC001
        --	for i in 1 to 100000 loop
        --            ld <= '1'; 
        --            addr <= x"C001";
        --            data <= x"0041"; wait for clock_period;
        --	end loop;
        --
        --	-- Now read a chanacter from UART port 0xC000
        --        ld <= '0'; 
        --        data <= (others => 'Z');
        --        addr <= x"C000"; wait for clock_period;
        --        
        wait for clock_period;
        report "memory test finished" severity failure;
    end process;
end;
