-------------------------------------------------------
-- Design Name : ram_sp_ar_sw
-- File Name   : ram_sp_ar_sw.vhd
-- Function    : Async read Sync write RAM 
-- Coder       : Deepak Kumar Tala (Verilog)
-- Translator  : Alexander H Pham (VHDL)
-------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ram is
    generic (ADDR_WIDTH :    integer := 8;
             DATA_WIDTH :    integer := 8
    );
    port (clk  :in    std_logic;                                -- Clock Input
        address:in    std_logic_vector(ADDR_WIDTH-1 downto 0);  -- address Input
        data   :inout std_logic_vector(DATA_WIDTH-1 downto 0);  -- data bi-directional
        cs     :in    std_logic;                                -- Chip Select
        we     :in    std_logic;                                -- Write Enable/Read Enable
        oe     :in    std_logic                                 -- Output Enable
      );
end entity;

architecture rtl of ram is
    ----------------Internal variables----------------
    constant RAM_DEPTH :integer := 2**ADDR_WIDTH;
    signal data_out :std_logic_vector (DATA_WIDTH-1 downto 0);
    type RAM is array (integer range <>)of
				std_logic_vector (DATA_WIDTH-1 downto 0);
    -- Initialise the memory to all zeroes
    signal mem : RAM (0 to RAM_DEPTH-1) := (others => (others => '0'));
begin
    ----------------Code Starts Here------------------
    -- Tri-State Buffer control
    -- output : When we = 0, oe = 1, cs = 1
    data <= data_out when (cs = '1' and oe = '1' and we = '0') else
	    (others=>'Z');
    
    -- Memory Write Block
    -- Write Operation : When we = 1, cs = 1
    MEM_WRITE:
      process (clk) begin
        if (rising_edge(clk)) then
            if (cs = '1' and we = '1') then
                mem(to_integer(unsigned(address))) <= data;
            end if;
        end if;
    end process;
    
    -- Memory Read Block
    -- Asynchronous Read Operation : When we = 0, oe = 1, cs = 1
    MEM_READ:
      process (address, cs, we, oe, mem) begin
        if (cs = '1' and we = '0' and oe = '1') then
            data_out <= mem(to_integer(unsigned(address)));
        end if;
    end process;
end architecture;

library ieee;
use ieee.std_logic_1164.all;

entity ram_test is
    generic (DATA_WIDTH :integer := 8;
             ADDR_WIDTH :integer := 8
    );
end ram_test;

architecture behaviour of ram_test is
    component ram is
       generic (ADDR_WIDTH :integer := ADDR_WIDTH;
                DATA_WIDTH :integer := DATA_WIDTH
       );
        port (
            clk     :in    std_logic;
            address :in    std_logic_vector (ADDR_WIDTH-1 downto 0);
            data    :inout std_logic_vector (DATA_WIDTH-1 downto 0);
            cs      :in    std_logic;
            we      :in    std_logic;
            oe      :in    std_logic);
    end component;
    
    signal clock  : std_logic;
    signal addr   : std_logic_vector(ADDR_WIDTH-1 downto 0);
    signal data   : std_logic_vector(DATA_WIDTH-1 downto 0);
    signal cs     : std_logic;
    signal we     : std_logic;
    signal oe     : std_logic;
    
    -- Clock period definitions
    constant clock_period : time := 10 ns;
    
begin
    -- Clock process definitions: clock with 50% duty cycle is generated here.
    clock_process: process
    begin
        clock <= '0';
        wait for clock_period/2;  --for 5 ns signal is '0'.
        clock <= '1';
        wait for clock_period/2;  --for next 5 ns signal is '1'.
    end process;
    
    -- Unit under test port map
    uut: ram port map (
        clk => clock,
        address => addr,
        data => data,
        cs => cs,
        we => we,
        oe => oe
      );
    
    --- Stimulation process
    stim_proc: process
    begin
        -- Read from RAM location 17
        cs <= '1'; 
        oe <= '1'; 
        we <= '0'; 
        addr <= x"017"; wait for clock_period;
        
        -- Store ab at RAM location 22
        cs <= '1'; 
        oe <= '1'; 
        we <= '1'; 
        addr <= x"022";
        data <= x"ab"; wait for clock_period;
        
        -- Store ee at RAM location 80
        cs <= '1'; 
        oe <= '1'; 
        we <= '1'; 
        addr <= x"080";
        data <= x"ee"; wait for clock_period;
        
        -- Read ab from RAM location 22
        cs <= '1'; 
        oe <= '1'; 
        we <= '0'; 
        data <= (others => 'Z');
        addr <= x"022"; wait for clock_period;
        
        -- Read from some other location
        cs <= '1'; 
        oe <= '1'; 
        we <= '0'; 
        data <= (others => 'Z');
        addr <= x"020"; wait for clock_period;
        
        -- Read from 22 again
        cs <= '1'; 
        oe <= '1'; 
        we <= '0'; 
        data <= (others => 'Z');
        addr <= x"022"; wait for clock_period;
        
        -- Read from 80 again
        cs <= '1'; 
        oe <= '1'; 
        we <= '0'; 
        data <= (others => 'Z');
        addr <= x"080"; wait for clock_period;
        wait for clock_period;
        
        report "ram test finished" severity failure;
    end process;
end;
